//
//  OpenGL widget
//
#include "ex12opengl.h"
#include <QtOpenGL>
#include <QMessageBox>
#include <math.h>

//
//  Constructor
//
Ex12opengl::Ex12opengl(QWidget* parent)
    : QGLWidget(parent)
{
   mode = 0;
   N = 1;
   init  = false;
   mouse = false;
   x0 = y0 = 0;
   zoom = 1;
   framebuf[0] = 0;
   framebuf[1] = 0;
}

//
//  Reset view
//
void Ex12opengl::reset()
{
   x0 = y0 = 0;
   zoom = 1;
   //  Request redisplay
   updateGL();
}

//
//  Set passes
//
void Ex12opengl::setPasses(int n)
{
   N = n;
   //  Request redisplay
   updateGL();
}

//
//  Set shader
//
void Ex12opengl::setShader(int sel)
{
   mode = sel;
   //  Request redisplay
   updateGL();
}

//
//  Set image
//
void Ex12opengl::setImage(QImage img)
{
   //  Copy image to texture
   glBindTexture(GL_TEXTURE_2D,frame);
   QImage rgba = QGLWidget::convertToGLFormat(img);
   w = rgba.width();
   h = rgba.height();
   glTexImage2D(GL_TEXTURE_2D,0,4,w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,rgba.bits());
   //  Allocate frame buffer objects to native camera resolution
   if (!framebuf[0] || framebuf[0]->width() != w|| framebuf[0]->height() != h)
   {
      for (int k=0;k<2;k++)
      {
         //  Allocate frame buffer objects
         if (framebuf[k]) delete framebuf[k];
         framebuf[k] = new QGLFramebufferObject(w,h);
      }
   }
   //  Request redisplay
   updateGL();
}

//
//  Initialize
//
void Ex12opengl::initializeGL()
{
   if (init) return;
   init = true;

   //  Load shaders
   Shader(shader[1],"",":/ex12a.frag");
   Shader(shader[2],"",":/ex12b.frag");
   Shader(shader[3],"",":/ex12c.frag");
   Shader(shader[4],"",":/ex12d.frag");
   Shader(shader[5],"",":/ex12e.frag");
   Shader(shader[6],"",":/ex12f.frag");
   Shader(shader[7],"",":/ex12g.frag");
   Shader(shader[8],"",":/ex12h.frag");

   //  Initialize texture
   glGenTextures(1,&frame);
   glBindTexture(GL_TEXTURE_2D,frame);
}

//
//  Projections are determined in paintGL
//
void Ex12opengl::resizeGL(int width, int height)
{
    Q_UNUSED(width);
    Q_UNUSED(height);
}

//
//  Draw the window
//
void Ex12opengl::paintGL()
{
   //  Starting texture
   glBindTexture(GL_TEXTURE_2D,frame);
   glEnable(GL_TEXTURE_2D);
   //  Camera aspect ratio
   float Casp = w / (float)h;

   //  Process
   if (mode)
   {
      //  Viewport is entire image
      glViewport(0,0,w,h);
      //  Set Projection
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(-Casp, +Casp, -1, +1, -1, +1);
      //  Set ModelView
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      //  Enable shader
      shader[mode].bind();

      //  Set shader increments
      shader[mode].setUniformValue("dX",(float)1.0/w);
      shader[mode].setUniformValue("dY",(float)1.0/h);

      //  Ping-Pong
      for (int k=0;k<N;k++)
      {
         //  Pick nearest pixel
         glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
         glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
         //  Set output buffer
         framebuf[k%2]->bind();
         //  Draw image
         glClear(GL_COLOR_BUFFER_BIT);
         glBegin(GL_QUADS);
         glTexCoord2f(0,0); glVertex2f(-Casp,-1);
         glTexCoord2f(1,0); glVertex2f(+Casp,-1);
         glTexCoord2f(1,1); glVertex2f(+Casp,+1);
         glTexCoord2f(0,1); glVertex2f(-Casp,+1);
         glEnd();
         //  Release output buffer
         framebuf[k%2]->release();
         //  Bind texture to result
         glBindTexture(GL_TEXTURE_2D,framebuf[k%2]->texture());
      }

      //  Release shader
      shader[mode].release();
   }

   //  Set pixel interpolation
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
   //  Window aspect ratio
   float Wasp = width() / (float)height();
   //  Viewport is entire window
   glViewport(0,0,width(),height());
   //  Set Projection
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(-Wasp, +Wasp, -1, +1, -1, +1);
   //  Set ModelView
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glScaled(zoom,zoom,1);
   glTranslated(x0,y0,0);
   //  Draw to screen
   glClear(GL_COLOR_BUFFER_BIT);
   glBegin(GL_QUADS);
   glTexCoord2f(0,0); glVertex2f(-Casp,-1);
   glTexCoord2f(1,0); glVertex2f(+Casp,-1);
   glTexCoord2f(1,1); glVertex2f(+Casp,+1);
   glTexCoord2f(0,1); glVertex2f(-Casp,+1);
   glEnd();
   //  Done with textures
   glDisable(GL_TEXTURE_2D);
}

//
//  Throw a fatal error and die
//
void Ex12opengl::Fatal(QString message)
{
   QMessageBox::critical(this,"Ex12opengl",message);
   QApplication::quit();
}

/******************************************************************/
/*************************  Mouse Events  *************************/
/******************************************************************/
//
//  Mouse pressed
//
void Ex12opengl::mousePressEvent(QMouseEvent* e)
{
   mouse = true;
   pos = e->pos();  //  Remember mouse location
}

//
//  Mouse released
//
void Ex12opengl::mouseReleaseEvent(QMouseEvent*)
{
    mouse = false;
}

//
//  Mouse moved
//
void Ex12opengl::mouseMoveEvent(QMouseEvent* e)
{
   if (mouse)
   {
      QPoint d = e->pos()-pos;      //  Change in mouse location
      x0 += d.x()/(zoom*width());   //  Translate x movement to azimuth
      y0 -= d.y()/(zoom*height());  //  Translate y movement to elevation
      pos = e->pos();               //  Remember new location
      updateGL();                   //  Request redisplay
   }
}

//
//  Mouse wheel
//
void Ex12opengl::wheelEvent(QWheelEvent* e)
{
   //  Zoom out
   if (e->delta()<0)
      zoom *= 1.05;
   //  Zoom in
   else if (zoom>1)
      zoom /= 1.05;
   //  Request redisplay
   updateGL();
}

//
//  Load shader
//
void Ex12opengl::Shader(QGLShaderProgram& shader,QString vert,QString frag)
{
   //  Vertex shader
   if (vert.length() && !shader.addShaderFromSourceFile(QGLShader::Vertex,vert))
      Fatal("Error compiling "+vert+"\n"+shader.log());
   //  Fragment shader
   if (frag.length() && !shader.addShaderFromSourceFile(QGLShader::Fragment,frag))
      Fatal("Error compiling "+frag+"\n"+shader.log());
   //  Link
   if (!shader.link())
      Fatal("Error linking shader\n"+shader.log());
}
